/*
* Author: Chris Campbell - www.iforce2d.net
*
* Copyright (c) 2006-2011 Erin Catto http://www.box2d.org
*
* This software is provided 'as-is', without any express or implied
* warranty.  In no event will the authors be held liable for any damages
* arising from the use of this software.
* Permission is granted to anyone to use this software for any purpose,
* including commercial applications, and to alter it and redistribute it
* freely, subject to the following restrictions:
* 1. The origin of this software must not be misrepresented; you must not
* claim that you wrote the original software. If you use this software
* in a product, an acknowledgment in the product documentation would be
* appreciated but is not required.
* 2. Altered source versions must be plainly marked as such, and must not be
* misrepresented as being the original software.
* 3. This notice may not be removed or altered from any source distribution.
*/

#ifndef IFORCE2D_WINDTUNNEL_H
#define IFORCE2D_WINDTUNNEL_H

#ifndef DEGTORAD
#define DEGTORAD 0.0174532925199432957f
#define RADTODEG 57.295779513082320876f
#endif

#include <set>
#include <vector>
#include <map>

#define AIR_PARTICLE_RADIUS 0.66666

int g_wingDesign = 0;
int g_nextWingDesign = 0;
int g_stepsPerPush = 5;

class Graph {
public:
    float dragPoints[180];
    float liftPoints[180];
    Graph() {
        for (int i = 0; i < 180; i++) dragPoints[i] = FLT_MAX;
        for (int i = 0; i < 180; i++) liftPoints[i] = FLT_MAX;
    }

    void addMeasurement(float angleDegrees, float drag, float lift) {
        int i = (int)(angleDegrees - -90);
        if ( i < 0 || i >= 180 )
            return;
        float uptake = dragPoints[i] == FLT_MAX ? 1 : 0.01f;
        dragPoints[i] = (1-uptake) * dragPoints[i] + uptake * drag;
        liftPoints[i] = (1-uptake) * liftPoints[i] + uptake * lift;
    }

    void draw( b2Vec2 offset, b2Vec2 scale ) {
        glEnable(GL_BLEND);
        glEnable(GL_POINT_SMOOTH);
        glPushMatrix();
        glTranslatef( offset.x, offset.y, 0 );
        glScalef( scale.x, scale.y, 1 );

        //frame
        glColor3f(0.75f, 0.75f, 0.75f);
        glBegin(GL_LINES);
        glVertex2f(-1, -1);
        glVertex2f(-1,  1);
        glVertex2f( 0, -1);
        glVertex2f( 0,  1);
        glVertex2f( 1, -1);
        glVertex2f( 1,  1);

        glVertex2f(-1, -1);
        glVertex2f( 1, -1);
        glVertex2f(-1,  0);
        glVertex2f( 1,  0);
        glVertex2f(-1,  1);
        glVertex2f( 1,  1);
        glEnd();


        glPointSize(8);
        glBegin(GL_POINTS);
        glColor3f(1,0,0);
        for (int i = 0; i < 179; i++) {
            if ( dragPoints[i] != FLT_MAX )
                glVertex2f( -1 + i * 2 * 1/180.0f, 0.2f * dragPoints[i] );
        }
        glColor3f(0,1,0);
        for (int i = 0; i < 179; i++) {
            if ( liftPoints[i] != FLT_MAX )
                glVertex2f( -1 + i * 2 * 1/180.0f, 0.2f * liftPoints[i] );
        }
        glEnd();

        glPopMatrix();
    }
};

Graph g_graphs[3];

#define MAX_TRACK_POSITIONS 10
class ParticleTrack {
public:
    b2Vec2 positions[MAX_TRACK_POSITIONS];
};

class iforce2d_windTunnel : public Test
{
public:
    iforce2d_windTunnel()
    {
        g_wingDesign = g_nextWingDesign;

        //include dumped world data
        if ( g_wingDesign == 2 ) {
            #include "windTunnel2.cpp"
        }
        else if ( g_wingDesign == 1 ) {
            #include "windTunnel1.cpp"
        }
        else  {
            #include "windTunnel0.cpp"
        }

        //find the pusher body (only kinematic body in the world)
        m_pusherBody = NULL;
        for (b2Body* b = m_world->GetBodyList(); b; b = b->GetNext()) {
            if ( b->GetType() == b2_kinematicBody ) {
                m_pusherBody = b;
                break;
            }
        }

        //find the catcher body (will be only body with position x > 19)
        m_catcherBody = NULL;
        for (b2Body* b = m_world->GetBodyList(); b; b = b->GetNext()) {
            if ( b->GetPosition().x > 19 ) {
                m_catcherBody = b;
                break;
            }
        }

        //find the wing body (will be dynamic body with mass over 50)
        m_wingBody = NULL;
        for (b2Body* b = m_world->GetBodyList(); b; b = b->GetNext()) {
            if ( b->GetMass() > 50 ) {
                m_wingBody = b;
                break;
            }
        }

        //m_wingBody->SetTransform( m_wingBody->GetPosition(), -m_angleOfAttack * DEGTORAD );

        b2MotorJointDef jd;
        jd.bodyA = m_groundBody;
        jd.bodyB = m_wingBody;
        jd.linearOffset = m_wingBody->GetPosition() - m_groundBody->GetPosition();
        jd.angularOffset = 0;
        jd.maxForce = 100000;
        jd.maxTorque = 100000;
        jd.correctionFactor = 1;
        m_controlJoint = (b2MotorJoint*)m_world->CreateJoint(&jd);

        m_stepsSinceLastColumn = 0;

        m_angleOfAttack = 0;
        m_drag = 0;
        m_lift = 0;
    }

    void addAirColumn()
    {
        float columnHeight = 40;
        int particlesPerColumn = columnHeight / (2 * AIR_PARTICLE_RADIUS);
        while ( m_queuedParticles.size() < particlesPerColumn ) {
            b2BodyDef bd;
            bd.type = b2_dynamicBody;
            bd.fixedRotation = true;
            bd.linearDamping = 1;
            bd.gravityScale = 0;
            bd.position = b2Vec2(-30, -5) + m_queuedParticles.size() * b2Vec2(3 * AIR_PARTICLE_RADIUS, 0);
            b2Body* body = m_world->CreateBody( &bd );

            b2CircleShape circleShape;
            circleShape.m_radius = AIR_PARTICLE_RADIUS;

            b2FixtureDef fd;
            fd.shape = &circleShape;
            fd.density = 0.1f;
            fd.friction = 0;
            fd.restitution = 0.0f;
            body->CreateFixture( &fd );

            m_queuedParticles.push_back(body);
        }
        for (int i = 0; i < particlesPerColumn; i++) {
            float height = AIR_PARTICLE_RADIUS + i * 2 * AIR_PARTICLE_RADIUS;
            b2Body* body = m_queuedParticles.back();
            body->SetTransform( b2Vec2( -30 + AIR_PARTICLE_RADIUS, height ), 0 );
            body->SetLinearVelocity( b2Vec2(0,0) );
            m_queuedParticles.pop_back();
            m_flyingParticles.insert(body);

            ParticleTrack track;
            for (int k = 0; k < MAX_TRACK_POSITIONS; k++)
                track.positions[k] = body->GetPosition();
            m_tracks[body] = track;
        }
    }

    void Keyboard(unsigned char key)
    {
        switch (key) {
        case 'a': g_nextWingDesign = (g_nextWingDesign+1) % 3; break;
        case 's': m_angleOfAttack -= 1; m_angleOfAttack = b2Clamp( m_angleOfAttack, -90.0f, 90.0f ); m_controlJoint->SetAngularOffset( -m_angleOfAttack * DEGTORAD ); break;
        case 'w': m_angleOfAttack += 1; m_angleOfAttack = b2Clamp( m_angleOfAttack, -90.0f, 90.0f ); m_controlJoint->SetAngularOffset( -m_angleOfAttack * DEGTORAD ); break;
        default: Test::Keyboard(key);
        }
    }

    void Step(Settings* settings)
    {
        Test::Step(settings);

        m_debugDraw.DrawString(5, m_textLine, "Press a to change wing design (change will take effect after restart: %d)", g_nextWingDesign);
        m_textLine += 15;
        m_debugDraw.DrawString(5, m_textLine, "Press s/w to change target angle of attack (currently: %.1f degrees)", m_angleOfAttack);
        m_textLine += 15;
        m_debugDraw.DrawString(5, m_textLine, "Current actual angle of attack: %.1f degrees", -m_wingBody->GetAngle() * RADTODEG);
        m_textLine += 15;

        m_debugDraw.DrawString(5, m_textLine, "Drag: %.2f", m_drag);
        m_textLine += 15;
        m_debugDraw.DrawString(5, m_textLine, "Lift: %.2f", m_lift);
        m_textLine += 15;

        if ( !settings->pause || settings->singleStep ) {
            for (int i = 0; i < m_toBeQueuedParticles.size(); i++) {
                b2Vec2 queuedPos = b2Vec2(-30, -50) + m_queuedParticles.size() * b2Vec2(3 * AIR_PARTICLE_RADIUS, 0);
                m_toBeQueuedParticles[i]->SetTransform( queuedPos, 0 );
                m_toBeQueuedParticles[i]->SetLinearVelocity( b2Vec2(0,0) );
                m_queuedParticles.push_back( m_toBeQueuedParticles[i] );
                m_flyingParticles.erase( m_toBeQueuedParticles[i] );
                map<b2Body*, ParticleTrack>::iterator it = m_tracks.find( m_toBeQueuedParticles[i] );
                if ( it != m_tracks.end() )
                    m_tracks.erase( it );
            }
            m_toBeQueuedParticles.clear();

            m_stepsSinceLastColumn++;
            if ( m_stepsSinceLastColumn > g_stepsPerPush ) {
                m_pusherBody->SetTransform( b2Vec2(-30,20), 0 );
                m_pusherBody->SetLinearVelocity( b2Vec2( 2*AIR_PARTICLE_RADIUS/(float)g_stepsPerPush*settings->hz, 0) );
                addAirColumn();
                m_stepsSinceLastColumn = 0;
            }

            float dragNow = -m_controlJoint->GetReactionForce(1/settings->hz).x;
            m_drag = 0.95f * m_drag + 0.05f * dragNow;
            float liftNow = -m_controlJoint->GetReactionForce(1/settings->hz).y;
            m_lift = 0.95f * m_lift + 0.05f * liftNow;
            g_graphs[g_wingDesign].addMeasurement( -m_wingBody->GetAngle() * RADTODEG, dragNow, liftNow );

            //update track points
            map<b2Body*, ParticleTrack>::iterator it = m_tracks.begin();
            map<b2Body*, ParticleTrack>::iterator end = m_tracks.end();
            while (it != end) {
                ParticleTrack& track = it->second;
                for (int k = MAX_TRACK_POSITIONS-2; k >= 0; k--)
                    track.positions[k+1] = track.positions[k];
                track.positions[0] = it->first->GetPosition();
                ++it;
            }
        }

        for (int i = 0; i < 3; i++)
            g_graphs[i].draw( b2Vec2(-25+i*25, 55), b2Vec2(10, 10) );

        glLineWidth(3);
        glBegin(GL_LINES);
        map<b2Body*, ParticleTrack>::iterator it = m_tracks.begin();
        map<b2Body*, ParticleTrack>::iterator end = m_tracks.end();
        while (it != end) {
            ParticleTrack& track = it->second;
            float dist = 0;
            for (int k = 0; k < MAX_TRACK_POSITIONS-1; k++)
                dist += (track.positions[k] - track.positions[k+1]).Length();
            //dist *= 0.5f;
            dist = b2Min(dist, 1.0f);
            glColor4f(1,dist*dist,0.5,0.75);
            for (int k = 0; k < MAX_TRACK_POSITIONS-1; k++) {
                glVertex2fv( (GLfloat*)&track.positions[k] );
                glVertex2fv( (GLfloat*)&track.positions[k+1] );
            }
            ++it;
        }
        glEnd();
        glLineWidth(1);
    }

    void BeginContact(b2Contact* contact)
    {
        b2Fixture* fixtureA = contact->GetFixtureA();
        b2Fixture* fixtureB = contact->GetFixtureB();

        if ( fixtureA->GetBody() == m_catcherBody && fixtureB->GetBody()->GetMass() < 10 ) {
            m_toBeQueuedParticles.push_back(fixtureB->GetBody());
            contact->SetEnabled(false);
        }
        else if ( fixtureB->GetBody() == m_catcherBody && fixtureA->GetBody()->GetMass() < 10 ) {
            m_toBeQueuedParticles.push_back(fixtureA->GetBody());
            contact->SetEnabled(false);
        }
    }

    static Test* Create()
    {
        return new iforce2d_windTunnel;
    }

    b2Body* m_pusherBody;
    b2Body* m_catcherBody;
    b2Body* m_wingBody;

    vector<b2Body*> m_queuedParticles;
    vector<b2Body*> m_toBeQueuedParticles;

    set<b2Body*> m_flyingParticles;
    map<b2Body*, ParticleTrack> m_tracks;

    int m_stepsSinceLastColumn;

    b2MotorJoint* m_controlJoint;
    float m_drag;
    float m_lift;

    float m_angleOfAttack;
};

#endif









